from __future__ import annotations
from abc import ABC, abstractmethod
from pyictacp.typez import ByteOrder
from pyictacp.checksum import Checksum
from pyictacp.encryption import Encryption
from typing import List
from pyictacp.packet.data import DataPacketData
from pyictacp.packet.command import CommandPacket

class Connection(ABC):
    def __init__(self, encryption: Encryption|None=None, checksum: Checksum|None=None, byteorder: ByteOrder='little') -> None:
        self.encryption = encryption
        self.checksum = checksum
        self.byteorder = byteorder
        
    @abstractmethod
    def execute_command(self, command: CommandPacket) -> List[DataPacketData] | None:
        pass



from __future__ import annotations
from abc import ABC, abstractclassmethod, abstractmethod
from pyictacp.typez import ByteOrder, NACKErrorCode
from pyictacp.packet import Packet

class SystemPacket(Packet, ABC, packet_type=0xC0):
    def __init_subclass__(cls) -> None:
        pass

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder) -> SystemPacket:
        if data[0] == 0xFF and data[1] == 0x00:
            return ACKPacket()
        elif data[0] == 0xFF and data[1] == 0xFF:
            return NACKPacket(
                NACKErrorCode(int.from_bytes(data[2:4], byteorder))
            )
        raise ValueError("Unknown packet type")

    def _to_bytes(self, byteorder: ByteOrder) -> bytes:
        if isinstance(self, ACKPacket):
            return bytes([0xFF, 0x00])
        elif isinstance(self, NACKPacket):
            return bytes([
                0xFF, 0xFF, #NACK
            ]) + self.error_code.value.to_bytes(2, byteorder)


class ACKPacket(SystemPacket):
    pass

class NACKPacket(SystemPacket):
    def __init__(self, error_code: NACKErrorCode) -> None:
        self.error_code = error_code


from pyictacp.connection import Connection
from pyictacp.record import Record
from pyictacp.packet.data import InputStatusDataPacketData
from pyictacp.packet.command import PermanentInputBypassCommandPacket, RemoveInputBypassCommandPacket, RequestInputStatusCommandPacket, TemporaryInputBypassCommandPacket

class Input(Record,
        data_class = InputStatusDataPacketData,
        data_request_command = RequestInputStatusCommandPacket,
        data_index_match = lambda rec, id: rec.input_index == id):
    
    def __init__(self, connection: Connection, record_id: int):
        super().__init__(connection, record_id)
        self.input_state = None
        self.bypassed = None
        self.bypassed_latched = None
        self.siren_lockout = None

    def _update(self, data: InputStatusDataPacketData):
        self.input_state = data.input_state
        self.bypassed = data.bypassed
        self.bypassed_latched = data.bypassed_latched
        self.siren_lockout = data.siren_lockout

    
    def remove_bypass(self):
        self.connection.execute_command(
            RemoveInputBypassCommandPacket(self.record_id)
        )

    def bypass(self, temporary: bool=False):
        cmd_type = TemporaryInputBypassCommandPacket if temporary else PermanentInputBypassCommandPacket

        self.connection.execute_command(
            cmd_type(self.record_id)
        )



from typing import Any, List, Type, TypeVar

Class = TypeVar('Class')

def all_subclasses(cls:Class) -> List[Class]:
    subclasses = []

    for scls in cls.__subclasses__():
        subclasses.append(scls)
        subclasses.extend(all_subclasses(scls))
    
    return subclasses


from abc import ABC, abstractmethod


class Checksum(ABC):
    length = 0
    def __init_subclass__(cls, length:int = 0) -> None:
        cls.length = length

    @abstractmethod
    def calculate(self, data: bytes) -> bytes:
        pass

class ByteChecksum(Checksum, length=1):    
    def calculate(self, data: bytes) -> bytes:
        return bytes([
            sum(data) % 256
        ])


#TODO: CRC16


from __future__ import annotations
from pyictacp.connection import Connection
from pyictacp.record import Record
from pyictacp.packet.data import AreaStatusDataPacketData
from pyictacp.packet.command import ArmAreaCommandPacket, Disarm24AreaCommandPacket, DisarmAllAreaCommandPacket, DisarmAreaCommandPacket, ForceArmAreaCommandPacket, InstantArmAreaCommandPacket, RequestAreaStatusCommandPacket, StayArmAreaCommandPacket


class Area(Record,
        data_class=AreaStatusDataPacketData,
        data_request_command=RequestAreaStatusCommandPacket,
        data_index_match=lambda rec, id: rec.area_index == id):
    
    def __init__(self, connection: Connection, record_id: int):
        super().__init__(connection, record_id)
        self.area_state = None
        self.area_tamper_state = None
        self.alarm_active = None
        self.siren_active = None
        self.alarm_in_memory = None
        self.remote_armed = None
        self.instant_armed = None
        self.partial_armed = None

    def _update(self, data: AreaStatusDataPacketData):
        assert data.area_index == self.record_id
        self.area_state = data.area_state
        self.area_tamper_state = data.area_tamper_state
        self.alarm_active = data.alarm_active
        self.siren_active = data.siren_active
        self.alarm_in_memory = data.alarm_in_memory
        self.remote_armed = data.remote_armed
        self.instant_armed = data.instant_armed
        self.partial_armed = data.partial_armed

    def arm(self):
        self.connection.execute_command(
            ArmAreaCommandPacket(self.record_id)
        )

    def force_arm(self):
        self.connection.execute_command(
            ForceArmAreaCommandPacket(self.record_id)
        )
    
    def stay_arm(self):
        self.connection.execute_command(
            StayArmAreaCommandPacket(self.record_id)
        )

    def instant_arm(self):
        self.connection.execute_command(
            InstantArmAreaCommandPacket(self.record_id)
        )

    def disarm(self):
        self.connection.execute_command(
            DisarmAreaCommandPacket(self.record_id)
        )
    
    def disarm_all(self):
        self.connection.execute_command(
            DisarmAllAreaCommandPacket(self.record_id)
        )

    def disarm_24hr(self):
        self.connection.execute_command(
            Disarm24AreaCommandPacket(self.record_id)
        )


from __future__ import annotations
from abc import ABC, abstractclassmethod, abstractmethod
from pyictacp.util import all_subclasses
from pyictacp.encryption import Encryption
from pyictacp.checksum import Checksum
from pyictacp.typez import ByteOrder

import logging
logger = logging.getLogger(__name__)

class Packet(ABC):
    HEADER_BINARY = bytes([
        0x49, 0x43
    ])

    HEADER_ASCII = bytes([
        0x49, 0x40
    ])

    PACKET_LENGTH_BYTES=2

    packet_type = -1

    def __init_subclass__(cls, packet_type: int) -> None:
        logger.debug(f"Registering packet type {packet_type:0{2}X} to {cls}")
        cls.packet_type = packet_type

    def to_bytes(self, byteorder: ByteOrder='little', ascii_mode: bool=False, encryption: Encryption|None=None, checksum:Checksum|None=None) -> bytes:
        """Packs the packet into its byte represenation
        """

        format_byte = 0
        encrypted_data = self._to_bytes()

        if encryption is not None:
            format_byte |= encryption.__class__.encryption_type & 3
            encrypted_data = encryption.encrypt(encrypted_data)

        checksum_length = checksum.__class__.length if checksum is not None else 0

        packet_length = sum([
            len(Packet.HEADER_BINARY ),
            Packet.PACKET_LENGTH_BYTES,
            2, # Packet Type & Format Byte
            len(encrypted_data),
            checksum_length
        ])

        checksum_value = bytes([])

        if checksum is not None:
            checksum_value = checksum.calculate(
                Packet.HEADER_BINARY  +
                packet_length.to_bytes(Packet.PACKET_LENGTH_BYTES, byteorder) +
                bytes([
                    self.__class__.packet_type,
                    format_byte
                ]) + 
                encrypted_data
            )

        message_body = (
            packet_length.to_bytes(Packet.PACKET_LENGTH_BYTES, byteorder) +
            bytes([
                self.__class__.packet_type,
                format_byte
            ]) + 
            encrypted_data +
            checksum_value
        )
        
        if ascii_mode:
            message_body = message_body.hex().upper().encode("ascii")

        return (Packet.HEADER_ASCII if ascii_mode else Packet.HEADER_BINARY) + message_body


    @classmethod
    def from_bytes(cls, data: bytes, byteorder: ByteOrder='little', encryption: Encryption|None=None, checksum: Checksum|None=None) -> Packet:
        """Constructs a packet from the given packet
        """
        # Validates header, checksum, length, decrypts then passes data to subclass

        # If in ASCII mode, first convert to binary-type
        logger.debug(f"Constructing packet from bytes: {data}")

        if data[:len(Packet.HEADER_ASCII)] == Packet.HEADER_ASCII:
            logger.debug("ASCII packet found, converting to binary")
            data = Packet.HEADER_BINARY + bytes.fromhex(data[len(Packet.HEADER_ASCII):].decode("ascii"))

        if data[:len(Packet.HEADER_BINARY)] != Packet.HEADER_BINARY:
            raise ValueError(f"Invalid packet header. got: '{data[:len(Packet.HEADER_BINARY)]}', want: {Packet.HEADER_BINARY}")

        logger.debug(f"Found header")
        checksum_length = 0

        if checksum is not None:
            checksum_length = checksum.__class__.length

            actual_checksum_value = data[-checksum_length:]
            checksum_value = checksum.calculate(data[:-checksum_length]) 

            if actual_checksum_value != checksum_value:
                raise ValueError(f"Invalid checksum {checksum}. got: {checksum_value}, want: {actual_checksum_value}")

            logger.debug(f"Found and validated checksum as {checksum_value}")

        packet_length = int.from_bytes(data[len(Packet.HEADER_BINARY):len(Packet.HEADER_BINARY)+Packet.PACKET_LENGTH_BYTES], byteorder)

        if packet_length != len(data):
            raise ValueError(f"Invalid packet length. got: {len(data)}, want: {packet_length}")

        packet_type = data[len(Packet.HEADER_BINARY)+Packet.PACKET_LENGTH_BYTES]
        format_byte = data[len(Packet.HEADER_BINARY)+Packet.PACKET_LENGTH_BYTES+1]

        logger.debug(f"Packet type: {packet_type:0{2}X}, Format: {format_byte:0{2}X}")

        if (format_byte & 4) > 0:
            raise ValueError("Addressed packets are not supported")

        encryption_type = format_byte & 3
        given_encryption_type = encryption.__class__.encryption_type if encryption is not None else 0

        if encryption_type != given_encryption_type:
            raise ValueError(f"Passed encryption type differs from that within packet. got: {given_encryption_type}, want: {encryption_type}")

        decrypted_data = data[len(Packet.HEADER_BINARY)+Packet.PACKET_LENGTH_BYTES+2:]

        if checksum_length > 0:
            decrypted_data = decrypted_data[:-checksum_length]

        if encryption is not None:
            decrypted_data = encryption.decrypt(decrypted_data)
            logger.debug(f"Decrypted packet")
        
        for scls in all_subclasses(cls):
            if scls.packet_type == packet_type:
                logger.debug(f"Found matching packet type: {scls}")
                return scls._from_bytes(decrypted_data, byteorder)

        raise ValueError(f"Packet type '{packet_type}' was unknown")

        

    @abstractmethod
    def _to_bytes(self, byteorder: ByteOrder='little') -> bytes:
        pass

    @abstractclassmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> Packet:
        pass


# Import so the classes are registered
# Has to be done at bottom so Packet is constructed
import pyictacp.packet.data
import pyictacp.packet.system
import pyictacp.packet.command



from __future__ import annotations
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from pyictacp.connection import Connection
from typing import Callable, List, Type
from pyictacp.packet.data import DataPacketData
from pyictacp.packet.command import RecordCommandPacket

class Record(ABC):
    data_class = DataPacketData
    data_request_command = RecordCommandPacket
    data_index_match = lambda rec, id: False

    def __init_subclass__(cls,
            data_class:Type[DataPacketData],
            data_request_command:Type[RecordCommandPacket],
            data_index_match: Callable[[DataPacketData, int], bool]
        ) -> None:

        cls.data_class = data_class
        cls.data_request_command = data_request_command
        cls.data_index_match = data_index_match
    
    def __init__(self, connection: Connection, record_id: int):
        self.record_id = record_id
        self.connection = connection
        self.last_updated: datetime = None
        

    def update(self):
        """Request the current state, and load onto the object
        """
        
        result = self.connection.execute_command(
            self.__class__.data_request_command(
                self.record_id
            )
        )
        if result is None:
            raise ValueError("Returned data was not data")

        data = None
        for component in result:
            if (isinstance(component, self.__class__.data_class)
                and self.__class__.data_index_match(component, self.record_id)):
                data = component

        if data is None:
            raise ValueError("No data was found")

        self._update(data)
        self.last_updated = datetime.utcnow()
        

    @abstractmethod
    def _update(self, data: DataPacketData):
        """Update the record with the given data
        """
        pass


from __future__ import annotations
from abc import ABC, abstractclassmethod, abstractmethod
from pyictacp.util import all_subclasses
from pyictacp.packet import Packet
from typing import List, Literal
from pyictacp.typez import AreaState, AreaTamperState, ByteOrder, DoorLockState, DoorState, FirmwareType, InputState, OutputState

import logging
logger = logging.getLogger(__name__)

class DataPacket(Packet, packet_type=0x01):
    def __init__(self, data_components: List[DataPacketData]):
        self.data_components = data_components
    
    def _to_bytes(self, byteorder: ByteOrder) -> bytes:
        data = bytearray()

        for component in self.data_components:
            if not isinstance(component, TerminatorDataPacketData):
                data.extend(component.to_bytes(bytearray))
        
        data.extend(
            TerminatorDataPacketData().to_bytes()
        )
        
        return bytes(data)

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder) -> DataPacket:
        components = []
        logger.debug(f"Constructing data packet from bytes {data}")

        while len(data) > 0:
            data_len = data[2]
            logger.debug(f"Found data type of length {data_len}")

            components.append(DataPacketData.from_bytes(
                data=data[:3+data_len],
                byteorder=byteorder
            ))
            
            data = data[3+data_len:]

        return cls(
            components
        )

        


class DataPacketData(ABC):
    data_type = -1
    def __init_subclass__(cls, data_type:int, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        logger.debug(f"Registed data type {data_type:0{4}X} to {cls}")
        cls.data_type = data_type


    @abstractmethod
    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        """Given the current packet, generate the payload porition
        """
        pass
    
    @abstractclassmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> DataPacketData:
        """Takes the payload part of data, and creates a new instance
        """
        pass

    @classmethod
    def from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> DataPacketData:
        """Takes the current packet, including data datatype and length, and creates a new DataPacketData
        """
        data_type = int.from_bytes(data[0:2], byteorder)
        data_length = data[2]
        data_payload = data[3:]

        assert data_length == len(data_payload)

        for scls in all_subclasses(cls):
            if scls.data_type == data_type:
                return scls._from_bytes(
                    data_payload,
                    byteorder
                );

        raise ValueError(
            f"Data Type {data_type:0{4}X} is unkownown" 
        )

    def to_bytes(self, byteorder: ByteOrder='little') -> bytes:
        """Given the current packet, convert it to the data portion of a data packet
        """
        payload = self._to_bytes(byteorder)
        data_type = self.__class__.data_type
        return (
            data_type.to_bytes(2, byteorder) +
            bytes([len(payload)]) + 
            payload
        )






class PanelSerialNumberDataPacketData(DataPacketData, data_type=0x0000):
    def __init__(self, serial_number: int) -> None:
        self.serial_number = serial_number
        
    
    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.serial_number.to_bytes(4, byteorder)

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> PanelSerialNumberDataPacketData:
        return cls(
            int.from_bytes(data[:4], byteorder)
        )


class PanelHardwareVersionDataPacketData(DataPacketData, data_type=0x0001):
    def __init__(self, hardware_version: int) -> None:
        self.hardware_version = hardware_version

    
    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.hardware_version.to_bytes(1, byteorder)

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> PanelHardwareVersionDataPacketData:
        return cls(
            data[0]
        )

class FirmwareTypeDataPacketData(DataPacketData, data_type=0x0002):
    def __init__(self, firmware_type: FirmwareType) -> None:
        self.firmware_type = firmware_type
    
    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.firmware_type.encode("ASCII")

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> FirmwareTypeDataPacketData:
        return cls(
            data[:2].decode("ASCII")
        )

class FirmwareVersionDataPacketData(DataPacketData, data_type=0x0003):
    def __init__(self, firmware_version: int) -> None:
        self.firmware_version = firmware_version

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.firmware_version.to_bytes(2, byteorder)
    
    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> FirmwareVersionDataPacketData:
        return cls(
            int.from_bytes(data[:2], byteorder)
        )

class FirmwareBuildDataPacketData(DataPacketData, data_type=0x0004):
    def __init__(self, firmware_build: int) -> None:
        self.firmware_build = firmware_build

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.firmware_build.to_bytes(2, byteorder)
    
    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> FirmwareBuildDataPacketData:
        return cls(
            int.from_bytes(data[:2], byteorder)
        )        

class DoorStatusDataPacketData(DataPacketData, data_type=0x0100):
    def __init__(self, door_index: int, door_lock_state: DoorLockState, door_state: DoorState) -> None:
        self.door_index = door_index
        self.door_lock_state = door_lock_state
        self.door_state = door_state

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return (
            self.door_index.to_bytes(4, byteorder) +
            bytes([
                self.door_lock_state.value,
                self.door_state.value,
                0,
                0
            ])
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> DoorStatusDataPacketData:
        return cls(
            int.from_bytes(data[:4], byteorder),
            DoorLockState(data[4]),
            DoorState(data[5])
        )
    
class AreaStatusDataPacketData(DataPacketData, data_type=0x0200):
    def __init__(self,
                area_index: int,
                area_state: AreaState,
                area_tamper_state: AreaTamperState,
                alarm_active: bool=False,
                siren_active: bool=False,
                alarm_in_memory: bool=False,
                remote_armed: bool=False,
                instant_armed: bool=False,
                partial_armed: bool=False
                ) -> None:
        self.area_index = area_index
        self.area_state = area_state
        self.area_tamper_state = area_tamper_state
        self.alarm_active = alarm_active
        self.siren_active = siren_active
        self.alarm_in_memory = alarm_in_memory
        self.remote_armed = remote_armed
        self.instant_armed = instant_armed
        self.partial_armed = partial_armed

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        flags = (
            (1 << 0) if self.alarm_active else 0 |
            (1 << 1) if self.siren_active else 0 |
            (1 << 2) if self.alarm_in_memory else 0 |
            (1 << 3) if self.remote_armed else 0 |
            (1 << 4) if self.instant_armed else 0 |
            (1 << 5) if self.partial_armed else 0 
        )

        return (
            self.area_index.to_bytes(4, byteorder) +
            bytes([
                self.area_state.value,
                self.area_tamper_state.value,
                flags,
                0
            ])
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> AreaStatusDataPacketData:
        alarm_active, siren_active, alarm_in_memory, remote_armed, instant_armed, partial_armed = [
            (data[6] & (1<<i)) > 0
            for i in range(6)
        ]

        return cls(
            int.from_bytes(data[:4], byteorder),
            AreaState(data[4]),
            AreaTamperState(data[5]),
            alarm_active,
            siren_active,
            alarm_in_memory,
            remote_armed,
            instant_armed,
            partial_armed
        )

class OutputStatusDataPacketData(DataPacketData, data_type=0x0300):
    def __init__(self, output_index: int, output_reference: str, output_state: OutputState) -> None:
        self.output_index = output_index
        self.output_reference = output_reference
        self.output_state = output_state

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return (
            self.output_index.to_bytes(4, byteorder) +
            self.output_reference[:8].encode("ASCII") +
            bytes([
                self.output_state.value,
                0,
                0,
                0
            ])
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> OutputStatusDataPacketData:
        return cls(
            int.from_bytes(data[:4], byteorder),
            data[4:12].decode("ASCII"),
            OutputState(data[12])
        )
        
class InputStatusDataPacketData(DataPacketData, data_type=0x0400):
    def __init__(self,
        input_index: int,
        input_reference: str,
        input_state: InputState,
        bypassed: bool=False,
        bypassed_latched: bool=False,
        siren_lockout: bool=False,
        ) -> None:
        
        self.input_index = input_index
        self.input_reference = input_reference
        self.input_state = input_state
        self.bypassed = bypassed
        self.bypassed_latched = bypassed_latched
        self.siren_lockout = siren_lockout

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        flags = (
            (1<<0) if self.bypassed else 0 |
            (1<<1) if self.bypassed_latched else 0 |
            (1<<3) if self.siren_lockout else 0
        )
        return (
            self.input_index.to_bytes(4, byteorder) +
            self.input_reference[:8].encode("ASCII") +
            bytes([
                self.input_state.value,
                flags,
                0,
                0
            ])
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> InputStatusDataPacketData:
        bypassed, bypassed_latched, _, siren_lockout = [
            (data[6] & (1<<i)) > 0
            for i in range(4)
        ]
        return cls(
            int.from_bytes(data[:4], byteorder),
            data[4:12].decode("ASCII"),
            InputState(data[12]),
            bypassed,
            bypassed_latched,
            siren_lockout
        )

class VariableStatusDataPacketData(DataPacketData, data_type=0x0500):
    def __init__(self,
        variable_index: int,
        variable_value: int
    ) -> None:
        self.variable_index = variable_index
        self.variable_value = variable_value

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return (
            self.variable_index.to_bytes(4, byteorder) +
            self.variable_value.to_bytes(2, byteorder)
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> VariableStatusDataPacketData:
        return cls(
            int.from_bytes(data[:4], byteorder),
            int.from_bytes(data[4:6], byteorder)
        )


class TroubleInputStatusDataPacketData(DataPacketData, data_type=0x0600):
    def __init__(self,
        trouble_input_index: int,
        trouble_input_reference: str,
        trouble_input_state: InputState,
        bypassed: bool=False,
        bypassed_latched: bool=False
        ) -> None:
        
        self.trouble_input_index = trouble_input_index
        self.trouble_input_reference = trouble_input_reference
        self.trouble_input_state = trouble_input_state
        self.bypassed = bypassed
        self.bypassed_latched = bypassed_latched

    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        flags = (
            (1<<0) if self.bypassed else 0 |
            (1<<1) if self.bypassed_latched else 0 
        )
        return (
            self.trouble_input_index.to_bytes(4, byteorder) +
            self.trouble_input_reference[:8].encode("ASCII") +
            bytes([
                self.trouble_input_state.value,
                flags,
                0,
                0
            ])
        )

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> TroubleInputStatusDataPacketData:
        bypassed, bypassed_latched = [
            (data[6] & (1<<i)) > 0
            for i in range(2)
        ]
        return cls(
            int.from_bytes(data[:4], byteorder),
            data[4:12].decode("ASCII"),
            InputState(data[12]),
            bypassed,
            bypassed_latched
        )

class HumanReadableEventDataPacketData(DataPacketData, data_type=0x0130):
    def __init__(self, event_message: str) -> None:
        self.event_message = event_message
    
    def _to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        return self.event_message.encode("ASCII") + bytes([0])

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder = 'little') -> HumanReadableEventDataPacketData:
        return cls(
            data[:-1].decode("ASCII")
        )

class TerminatorDataPacketData(DataPacketData, data_type=0xFFFF):
    def _to_bytes(self, byteorder: ByteOrder) -> bytes:
        return bytes()
    
    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder) -> TerminatorDataPacketData:
        return cls()


if __name__ == "__main__":
    # Test we can resolve the bytes to a type

    org = bytes([
        0x00, 0x03,
        0x10,
        0x05, 0x00, 0x00, 0x00,
        0x43, 0x50, 0x30, 0x30, 0x31, 0x3A, 0x30, 0x35,
        0x01,
        0x00, 0x00, 0x00
    ])

    x = DataPacketData.from_bytes(org)

    print(x.output_index)
    print(x.output_reference)
    print(x.output_state)

    assert org == x.to_bytes()


from abc import ABC, abstractmethod


class Encryption(ABC):
    encryption_type=0
    def __init_subclass__(cls, encryption_type:int) -> None:
        cls.encryption_type = encryption_type

    @abstractmethod
    def encrypt(self, data: bytes) -> bytes:
        pass

    @abstractmethod
    def decrypt(self, data: bytes) -> bytes:
        pass

#TODO: Encrpytion (AES)



from __future__ import annotations
from abc import ABC, abstractmethod

from datetime import timedelta
from pyictacp.connection import Connection
from pyictacp.checksum import Checksum
from pyictacp.encryption import Encryption
from typing import List
from pyictacp.typez import ByteOrder, PanelInformation
from pyictacp.packet.data import DataPacketData, FirmwareBuildDataPacketData, FirmwareTypeDataPacketData, FirmwareVersionDataPacketData, PanelHardwareVersionDataPacketData, PanelSerialNumberDataPacketData
from pyictacp.packet.command import CommandPacket, InactivityTimeCommandPacket, LoginCommandPacket, LogoutCommandPacket, PanelDescriptionCommandPacket, PollCommandPacket
from pyictacp.record.area import Area
from pyictacp.record.door import Door
from pyictacp.record.input import Input
from pyictacp.record.output import Output
from pyictacp.record.troubleinput import TroubleInput
from pyictacp.record.variable import Variable


class ICTACPClient(ABC):
    def __init__(self, connection: Connection):
        self.connection = connection
        self.logged_in = False
        self.pin = ""

    

    def poll(self) -> bool:
        """Polls the controller
        """
        self.connection.execute_command(
            PollCommandPacket(),
            requires_login=False
        )
        
    def panel_description(self) -> PanelInformation:
        """Returns serial number, hardware version, firmware type, version and build
        """
        res = self.connection.execute_command(
            PanelDescriptionCommandPacket()
        )

        info: PanelInformation = {}

        if res is not None:
            for component in res:
                if isinstance(component, PanelSerialNumberDataPacketData):
                    info['serial_number'] = component.serial_number
                elif isinstance(component, PanelHardwareVersionDataPacketData):
                    info['hardware_version'] = component.hardware_version
                elif isinstance(component, FirmwareTypeDataPacketData):
                    info['firmware_type'] = component.firmware_type
                elif isinstance(component, FirmwareVersionDataPacketData):
                    info['firmware_version'] = component.firmware_version
                elif isinstance(component, FirmwareBuildDataPacketData):
                    info['firmware_build'] = component.firmware_build

            return info
            
        
        raise ValueError("ACK Packet was returned whilst expecting data")


    def login(self, pin_number: str):
        """Attempts a login to the controller
        """
        
        # Will raise exception if invalid login
        self.connection.execute_command(
            LoginCommandPacket(pin_number)
        )

        self.pin = pin_number
        self.logged_in = True

    def logout(self):
        """Logs out from the controller
        """

        if self.logged_in:
            # No use logging out if we aren't logged in
            self.connection.execute_command(
                LogoutCommandPacket()
            )
            self.logged_in = False
            self.pin = ""

    def inactivity_timeout(self, timespan: timedelta):
        """Set the inactivity timeout
        """
        total_seconds = int(timespan.total_seconds())

        if timespan > timedelta(minutes=100):
            raise ValueError("Timespan cannot be larger than 100 minutes")

        self.connection.execute_command(
            InactivityTimeCommandPacket(
                total_seconds
            )
        )

    #TODO: Monitor
    #TODO: Events

    def get_area(self, area_id: int) -> Area:
        area = Area(connection=self.connection, record_id=area_id)
        area.update()
        return area
    
    def get_door(self, door_id: int) -> Door:
        door = Door(connection=self.connection, record_id=door_id)
        door.update()
        return door

    def get_input(self, input_id: int) -> Input:
        inp = Input(connection=self.connection, record_id=input_id)
        inp.update()
        return inp

    def get_output(self, output_id: int) -> Output:
        out = Output(connection=self.connection, record_id=output_id)
        out.update()
        return out

    def get_trouble_input(self, trouble_input_id: int) -> TroubleInput:
        trouble_inp = TroubleInput(connection=self.connection, record_id=trouble_input_id)
        trouble_inp.update()
        return trouble_inp

    def get_variable(self, variable_id: int) -> Variable:
        var = Variable(connection=self.connection, record_id=variable_id)
        var.update()
        return var




    



from pyictacp.connection import Connection
from pyictacp.record import Record
from pyictacp.packet.data import VariableStatusDataPacketData
from pyictacp.packet.command import GetVariableCommandPacket, SetVariableCommandPacket

class Variable(Record,
        data_class = VariableStatusDataPacketData,
        data_request_command = GetVariableCommandPacket,
        data_index_match = lambda rec, id: rec.variable_index == id):

    def __init__(self, connection: Connection, record_id: int):
        super().__init__(connection, record_id)
        self.variable_value = None

    def set_value(self, new_value: int):
        if new_value > 0xFFFF:
            raise ValueError(f"new value is larger than {0xFFFF}")

        self.connection.execute_command(
            SetVariableCommandPacket(
                self.record_id,
                new_value
            )
        )

    def __get__(self, instance, owner) -> int:
        self.update()
        return self.variable_value
    
    def __set__(self, instance, new_value:int):
        self.set_value(
            new_value
        )







from pyictacp.connection import Connection
from pyictacp.record import Record
from pyictacp.packet.data import DoorStatusDataPacketData
from pyictacp.packet.command import RequestDoorStatusCommandPacket, LockDoorCommandPacket, UnlockDoorCommandPacket, UnlockDoorLatchedCommandPacket



class Door(Record,
        data_class=DoorStatusDataPacketData,
        data_request_command=RequestDoorStatusCommandPacket,
        data_index_match=lambda rec, id: rec.door_imdex == id
        ):

    def __init__(self, connection: Connection, record_id: int):
        super().__init__(connection, record_id)
        self.lock_state = None
        self.door_state = None

    def lock(self):
        self.connection.execute_command(
            LockDoorCommandPacket(
                self.record_id
            )
        )

    def unlock(self, latched:bool = False):
        cmd_type = UnlockDoorLatchedCommandPacket if latched else UnlockDoorCommandPacket

        self.connection.execute_command(
            cmd_type(
                self.record_id
            )
        )
    
    
    def _update(self, data: DoorStatusDataPacketData):
        assert data.door_index == self.record_id
        self.lock_state = data.door_lock_state
        self.door_state = data.door_state



from enum import Enum
from typing import Literal, TypedDict

ByteOrder = Literal['little', 'big']


class RecordType(Enum):
    ALL             = 0x0000
    DOOR            = 0x0001
    AREA            = 0x0002
    AREA_INPUT      = 0x0102
    OUTPUT          = 0x0003
    INPUT           = 0x0004
    VARIABLE        = 0x0005
    TROUBLE_INPUT   = 0x0006

class DoorLockState(Enum):
    LOCKED = 0x00
    UNLOCKED_ACCESS = 0x01
    UNLOCKED_SCHEDULE = 0x02
    UNLOCKED_TIMED = 0x03
    UNLOCKED_LATCHED = 0x04
    UNLOCKED_REX = 0x05
    UNLOCKED_REN = 0x06
    UNLOCKED_KEYPAD = 0x07
    UNLOCKED_AREA = 0x08
    UNLOCKED_FIRE = 0x09

class DoorState(Enum):
    CLOSED = 0x00
    OPEN = 0x01
    OPEN_ALERT = 0x02
    LEFT_OPEN = 0x03
    FORCED_OPEN = 0x04

class AreaState(Enum):
    DISARMED = 0x00
    INPUTS_OPEN = 0x01
    TROUBLE_CONDITION = 0x02
    BYPASS_ERROR = 0x03
    BYPASS_WARNING = 0x04
    NOT_VACANT = 0x05
    ARMED = 0x80
    EXIT_DELAY = 0x81
    ENTRY_DELAY = 0x82
    DISARM_DELAY = 0x83
    CODE_DELAY = 0x84

class AreaTamperState(Enum):
    DISARMED = 0x00
    BUSY = 0x01
    ARMED = 0x80

class OutputState(Enum):
    OFF = 0x00
    ON = 0x01
    ON_PULSED = 0x02
    ON_TIMED = 0x03
    ON_PULSED_TIMED = 0x04

class InputState(Enum):
    CLOSED = 0x00
    OPEN = 0x01
    SHORT = 0x02
    TAMPER = 0x03

class NACKErrorCode(Enum):
    SERVICE_INDEX_NOT_VALID = 0x0121
    SERVICE_COMMAND_NOT_VALID = 0x0120
    USER_LOGIN = 0x0300
    USER_LOGOUT = 0x0301
    USER_INVALID = 0x0302
    USER_AXS_AREA = 0x0303
    USER_DOOR_GROUP = 0x030A
    USER_AXS_DOOR_AXS_LVL = 0x030F
    DOOR_SVC_DENIED_LOCKDOWN = 0x0A23
    DOOR_ALREADY_IN_STATE = 0x0A32
    DOOR_INTERLOCK_ACTIVE = 0x0A12
    AREA_NO_CHANGE = 0x0869
    USER_ACCESS_RIGHTS = 0x0303
    INPUT_COMMAND_FAILED = 0x040E

FirmwareType = Literal['GX','SE','LE']

class PanelInformation(TypedDict):
    serial_number: str
    hardware_version: int
    firmware_type: FirmwareType
    firmware_version: int
    firmware_build: int


from __future__ import annotations
from pyictacp.checksum import Checksum
from pyictacp.typez import ByteOrder, NACKErrorCode
from pyictacp.packet.system import ACKPacket, NACKPacket
from typing import List
from pyictacp.packet.data import DataPacket, DataPacketData
from pyictacp.packet.command import CommandPacket
from pyictacp.packet import Packet
import socket
from pyictacp.encryption import Encryption
from pyictacp.connection import Connection

import logging
logger = logging.getLogger(__name__)

class TCPConnection(Connection):
    def __init__(self, host: str, port: int, encryption: Encryption|None=None, checksum: Checksum|None=None, byteorder: ByteOrder='little'):
        super().__init__(encryption=encryption, checksum=checksum, byteorder=byteorder)
        self.socket_addr = (host, port)
        self.socket = None


    def connect(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect(self.socket_addr)

    def close(self):
        self.socket.close()
        self.socket = None

    def _send(self, packet: bytes):
        self.socket.send(packet)

    def _recv(self) -> bytes:
        buf = bytearray()

        while True:
            if len(buf) < len(Packet.HEADER_BINARY):
                buf.extend(self.socket.recv(1))
            else:
                # We have enough bytes for the header, now check it
                is_valid = True
                for i, char in enumerate(Packet.HEADER_BINARY):
                    is_valid = is_valid and (char == buf[i])
                
                if is_valid:
                    # We found our header, now lets go to the length loop.
                    break
                buf = buf[1:] # Just cut the first byte off and try again

        length_bytes = self.socket.recv(2)
        buf.extend(length_bytes)
        expected_length = int.from_bytes(
            length_bytes,
            self.byteorder
        )

        while len(buf) != expected_length:
            # We have to do this to ensure we get the right length, and not less.
            buf.extend(self.socket.recv(1)) 

        
        return bytes(buf)


    def execute_command(self, command: CommandPacket) -> List[DataPacketData] | None:
        logger.debug(f"Transmitting packet: {command}")
        command_bytes = command.to_bytes(self.byteorder, False, self.encryption, self.checksum)
        
        logger.debug(f"Writing bytes: {command_bytes}")
        self._send(command_bytes)

        response_bytes = self._recv()
        logger.debug(f"Got bytes: {response_bytes}")
        response = Packet.from_bytes(response_bytes, self.byteorder, self.encryption, self.checksum)
        logger.debug(f"Recieving packet: {response}")


        if isinstance(response, ACKPacket):
            # No data, but we ACK'd fine
            return None 
        elif isinstance(response, NACKPacket):
            # We got a NACK, lets raise an error for this

            # First, check if its just because we aren't logged in
            # If our local state says we are, we probably timed out and can reauthenticate and try again
            if response.error_code == NACKErrorCode.USER_LOGOUT and self.logged_in:
                self.login(self.pin)
                return self.execute_command(command)

            raise ValueError(f"NACK packet: {response.error_code}") #TODO: Refactor errors
        elif issubclass(type(response), DataPacket):
            return response.data_components
        else:
            raise ValueError(f"Got type {type(response)} which is unhandled")




from __future__ import annotations
from pyictacp.connection import Connection
from pyictacp.record import Record
from pyictacp.packet.data import OutputStatusDataPacketData
from pyictacp.packet.command import RequestOutputStatusCommandPacket, OutputOffCommandPacket, OutputOnTimedCommandPacket, OutputOnCommandPacket
from datetime import timedelta

class Output(Record,
        data_class = OutputStatusDataPacketData,
        data_request_command = RequestOutputStatusCommandPacket,
        data_index_match=lambda rec, id: rec.output_index == id):

    def __init__(self, connection: Connection, record_id: int):
        super().__init__(connection, record_id)
        self.output_state = None

    def _update(self, data: OutputStatusDataPacketData):
        assert data.output_index == self.record_id
        self.output_state = data.output_state
    
    def off(self):
        self.connection.execute_command(
            OutputOffCommandPacket(self.record_id)
        )

    def on(self, time: timedelta | None = None):
        if time is not None:
            time_seconds = time.total_seconds()
            if time_seconds > 0xFFFF:
                raise ValueError(f"Time cannot exceed {0xFFFF} total seconds")

            self.connection.execute_command(
                OutputOnTimedCommandPacket(self.record_id, time_seconds)
            )
        else:
            self.connection.execute_command(
                OutputOnCommandPacket(self.record_id)
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyictacp', 'pyictacp.connection', 'pyictacp.packet', 'pyictacp.record']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyictacp',
    'version': '0.2.2',
    'description': 'ICT Protege Automation and Control Protocol wrapper',
    'long_description': None,
    'author': 'Thomas Hobson',
    'author_email': 'git@hexf.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from abc import ABC, abstractclassmethod, abstractmethod
from pyictacp.util import all_subclasses
from pyictacp.typez import ByteOrder, RecordType
from pyictacp.packet import Packet

import logging
logger = logging.getLogger(__name__)

class CommandPacket(Packet, ABC, packet_type=0x00):
    command_group = -1
    sub_command = -1

    def __init_subclass__(cls, command_group: int=None, sub_command: int=None) -> None:
        logging.debug(f"Registed command {cls} with id {command_group:0{2}X}:{sub_command:0{2}X}")
        cls.command_group = command_group
        cls.sub_command = sub_command

    def _to_bytes(self, byteorder: ByteOrder='little') -> bytes:
        return bytes([
            self.__class__.command_group,
            self.__class__.sub_command
        ]) + self._to_bytes_()

    @classmethod
    def _from_bytes(cls, data: bytes, byteorder: ByteOrder='little') -> CommandPacket:
        command_group = data[0]
        sub_command = data[1]

        payload = data[2:]
        for scls in all_subclasses(cls):
            if scls.command_group == command_group and scls.sub_command == sub_command:
                return scls._from_bytes_(payload, byteorder)

        raise ValueError(f"Command {command_group:X}:{sub_command:X} is unknown")
    
    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return bytes([])
    
    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> CommandPacket:
        return cls()

class RecordCommandPacket(CommandPacket, command_group=-1, sub_command=-1):
    def __init__(self, record_id: int):
        self.record_id = record_id

    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return self.record_id.to_bytes(4, byteorder)

    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> RecordCommandPacket:
        return cls(
            record_id = int.from_bytes(data[:4], byteorder)
        )


class PollCommandPacket(CommandPacket, command_group=0x00, sub_command=0x00):
    pass

class PanelDescriptionCommandPacket(CommandPacket, command_group=0x00, sub_command=0x01):
    pass

class LoginCommandPacket(CommandPacket, command_group=0x00, sub_command=0x02):
    MAX_PIN_LENGTH = 6

    def __init__(self, pin_number):
        self.pin_number = pin_number

        if len(pin_number) > LoginCommandPacket.MAX_PIN_LENGTH:
            raise ValueError("Provided pin has too many digits")

    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        pin_digits = map(int, self.pin_number)
        pin_bytes = bytes(pin_digits)

        logger.debug(f"Login pin, pin: {self.pin_number} digits: {[*pin_digits]}")
        

        terminator = bytes([])

        if len(pin_bytes) < LoginCommandPacket.MAX_PIN_LENGTH:
            terminator = bytes([255])

        return pin_bytes + terminator

    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> LoginCommandPacket:
        pin = data[:LoginCommandPacket.MAX_PIN_LENGTH]

        if pin[-1] not in [
            0x0F, # Shift In
            0x18, # Cancel
            0x0D, # Carriage return
            ] + range(10):
            # This is a terminator
            pin = pin[:-1]

        pin_digits = map(str, pin)

        return cls(
            pin_digits
        )

class LogoutCommandPacket(CommandPacket, command_group=0x00, sub_command=0x03):
    pass

class InactivityTimeCommandPacket(CommandPacket, command_group=0x00, sub_command=0x04):
    def __init__(self, inactivity_seconds: int):
        self.inactivity_time = inactivity_seconds
        if inactivity_seconds > 6000: # 100 minutes is max
            raise ValueError("Maximum inactivity time is 100 minutes")
    
    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return self.inactivity_time.to_bytes(2, byteorder)

    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> InactivityTimeCommandPacket:
        return cls(
            int.from_bytes(data[:2], byteorder)
        )

class MonitorCommandPacket(CommandPacket, command_group=0x00, sub_command=0x05):
    def __init__(self, record_type: RecordType, record_id: int, start_monitoring:bool = True, force_update: bool = False):
        self.record_type = record_type
        self.record_id = record_id
        self.start_monitoring = start_monitoring
        self.force_update = force_update

        if start_monitoring and record_type is RecordType.ALL:
            raise ValueError("Cannot monitor ALL records")
        
        if not start_monitoring and force_update:
            raise ValueError("Cannot force-update when stopping monitoring")

    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        flags = 0
        if self.start_monitoring:
            flags |= 1
        if self.force_update:
            flags |= 2

        return (
            self.record_type.value.to_bytes(2, byteorder) + 
            self.record_id.to_bytes(4, byteorder) +
            bytes([flags, 0]),
        )
    
    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> MonitorCommandPacket:
        return cls(
            RecordType(int.from_bytes(data[2:], byteorder)),
            int.from_bytes(data[2:6], byteorder),
            data[6]&1 > 0,
            data[6]&2 > 0
        )


class EventsCommandPacket(CommandPacket, command_group=0x00, sub_command=0x06):
    def __init__(self, start_events=True, numerical_form=True, send_immediately=True):       
        self.start_events = start_events
        self.numerial_form = numerical_form
        self.send_immediately = send_immediately

    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return bytes([
            1 if self.start_events else 0,
            (
                0 if self.numerial_form else 1 |
                0 if self.send_immediately else 2
            )
        ])
    
    @classmethod
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> EventsCommandPacket:
        return cls(
            data[0] == 1,
            data[1]&1 > 0,
            data[1]&2 > 0
        )
        

# 0x00 0x07 - ACK Configuration, ignored as changes protocol
class LockDoorCommandPacket(RecordCommandPacket, command_group=0x01, sub_command=0x00):
    pass

class UnlockDoorCommandPacket(RecordCommandPacket, command_group=0x01, sub_command=0x01):
    pass

class UnlockDoorLatchedCommandPacket(RecordCommandPacket, command_group=0x01, sub_command=0x02):
    pass

class RequestDoorStatusCommandPacket(RecordCommandPacket, command_group=0x01, sub_command=0x80):
    pass

class DisarmAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x00):
    pass

class Disarm24AreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x01):
    pass

class DisarmAllAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x02):
    pass

class ArmAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x03):
    pass

class ForceArmAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x04):
    pass

class StayArmAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x05):
    pass

class InstantArmAreaCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x06):
    pass

class RequestAreaStatusCommandPacket(RecordCommandPacket, command_group=0x02, sub_command=0x80):
    pass

class OutputOffCommandPacket(RecordCommandPacket, command_group=0x03, sub_command=0x00):
    pass

class OutputOnCommandPacket(RecordCommandPacket, command_group=0x03, sub_command=0x01):
    pass

class OutputOnTimedCommandPacket(RecordCommandPacket, command_group=0x03, sub_command=0x02):
    def __init__(self, record_id: int, on_time_seconds: int):
        super().__init__(record_id)
        self.on_time = on_time_seconds
    
    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return super()._to_bytes_(byteorder) + self.on_time.to_bytes(2, byteorder)
    
    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder='little') -> OutputOnTimedCommandPacket:
        return cls(
            int.from_bytes(data[:4], byteorder),
            int.from_bytes(data[2:6], byteorder),
        )


class RequestOutputStatusCommandPacket(RecordCommandPacket, command_group=0x03, sub_command=0x80):
    pass

class RemoveInputBypassCommandPacket(RecordCommandPacket, command_group=0x04, sub_command=0x00):
    pass

class TemporaryInputBypassCommandPacket(RecordCommandPacket, command_group=0x04, sub_command=0x01):
    pass

class PermanentInputBypassCommandPacket(RecordCommandPacket, command_group=0x04, sub_command=0x02):
    pass

class RequestInputStatusCommandPacket(RecordCommandPacket, command_group=0x04, sub_command=0x80):
    pass

class SetVariableCommandPacket(RecordCommandPacket, command_group=0x05, sub_command=0x00):
    def __init__(self, record_id: int, new_value: int):
        super().__init__(record_id)
        self.new_value = new_value

    def _to_bytes_(self, byteorder: ByteOrder='little') -> bytes:
        return super()._to_bytes_() + self.new_value.to_bytes(2, byteorder)

    def _from_bytes_(cls, data: bytes, byteorder: ByteOrder) -> SetVariableCommandPacket:
        return cls(
            int.from_bytes(data[:4], byteorder),
            int.from_bytes(data[2:6], byteorder),
        )


class GetVariableCommandPacket(RecordCommandPacket, command_group=0x04, sub_command=0x80):
    pass

class RemoveTroubleInputBypassCommandPacket(RecordCommandPacket, command_group=0x06, sub_command=0x00):
    pass

class TemporaryTroubleInputBypassCommandPacket(RecordCommandPacket, command_group=0x06, sub_command=0x01):
    pass

class PermanentTroubleInputBypassCommandPacket(RecordCommandPacket, command_group=0x06, sub_command=0x02):
    pass

class RequestTroubleInputStatusCommandPacket(RecordCommandPacket, command_group=0x06, sub_command=0x80):
    pass


if __name__ == "__main__":
    pkt_bytes = bytes([
        0x49, 0x43, # Header
        0x0C, 0x00, # Length
        0x00,       # Command packet 
        0x00,       # No encryption
        0x01, 0x01, # Unlock door command
        0x07, 0x00, 0x00, 0x00 # index 7
    ])
    pkt = Packet.from_bytes(pkt_bytes)
    
    print(pkt)

    assert pkt.to_bytes() == pkt_bytes